

#include "SchedulerExample.h"
#include "SimpleTime.h"


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef min
#undef max

#include <iostream>


template<> SchedulerExample* Singleton<SchedulerExample>::ms_Singleton = 0;



SchedulerExample::SchedulerExample() :
	_root(0),
	_sceneManager(0),
	_camera(0),
	_inputManager(0),
	_keyboard(0),
	_continueRendering(true)
{
	Time::initialise();

	// variable queue uses a moving frame average to decide the cutoff frame length
	_movingAverageTaskQueue.setMovingAverageFrames(100);

	_fixedLimitQueue.setFrameLimit(20);
}

bool SchedulerExample::initialise()
{
	// Ogre initialisation
	// use the right configuration file depending on build mode
#ifdef _DEBUG
	std::string configFilename("ogre_d.cfg");
#else
	std::string configFilename("ogre.cfg");
#endif

	// initialise root
	_root = new Root("plugins.cfg", configFilename);

	// use current config if present, otherwise show cfg dialog
	if(_root->restoreConfig() == false)
	{
		if(_root->showConfigDialog() == false) return false;
	}

	RenderWindow* renderWindow = _root->initialise(true, "Task Scheduler Example");
	_sceneManager = _root->getSceneManager(ST_GENERIC);

	// this shouldn't be neccessary - what am I doing wrong?
	_sceneManager->_setDestinationRenderSystem(_root->getRenderSystem());

	// resources 
	ResourceGroupManager::getSingleton().addResourceLocation("data", "FileSystem");
	ResourceGroupManager::getSingleton().initialiseResourceGroup(ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
	ResourceGroupManager::getSingleton().loadResourceGroup(ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, true, false);

	// create camera and viewport
	_camera = _sceneManager->createCamera("DefaultCamera");
	Viewport* viewport = renderWindow->addViewport(_camera, 0);
	viewport->setBackgroundColour(ColourValue::Black);

	// initialise OIS input handling
	OIS::ParamList pl;	
	size_t windowHnd = 0;
	std::stringstream windowHndStr;

#if defined OIS_WIN32_PLATFORM
	renderWindow->getCustomAttribute("HWND", &windowHnd);
#elif defined OIS_LINUX_PLATFORM
	renderWindow->getCustomAttribute("GLXWINDOW", &windowHnd);
#endif
	//Both x11 and Win32 use handle of some sort..
	windowHndStr << windowHnd;
	pl.insert(std::make_pair(std::string("WINDOW"), windowHndStr.str()));

	_inputManager = OIS::InputManager::createInputSystem(pl);

	bool bufferInput = true;
	_keyboard = static_cast<OIS::Keyboard*>(_inputManager->createInputObject(OIS::OISKeyboard, bufferInput));
	_keyboard->setEventCallback(this);

	// frame-locked tasks
	_frameLockedQueue.addTask(MakeDelegate(this, &SchedulerExample::renderFrame), 1.0f);
	_frameLockedQueue.addTask(MakeDelegate(this, &SchedulerExample::frameRate), 0.0f);
	_frameLockedQueue.addTask(MakeDelegate(this, &SchedulerExample::captureInput), 0.1f);

	return true;
}

void SchedulerExample::mainLoop()
{
	unsigned long lastMS = Time::getMilliseconds();
	while(_continueRendering)
	{
		// update time, find elapsed time since last frame
		unsigned long nowMS = Time::getMilliseconds();
		unsigned long elapsedMS = nowMS - lastMS;
		lastMS = nowMS;

		// run frame-locked tasks
		_frameLockedQueue.step(elapsedMS);

		// run variable tasks
		_movingAverageTaskQueue.step(elapsedMS);
		_fixedLimitQueue.step(elapsedMS);
	}
}


void SchedulerExample::subscribeKeyboard(const std::string& identifier, KeyboardCallback callback)
{
	_keyboardSubscribers[identifier] = callback;
}

void SchedulerExample::unsubscribeKeyboard(const std::string& identifier)
{
	_keyboardSubscribers.erase(identifier);
/*
	SchedulerExample::KeyboardSubscriberMap::iterator found = _keyboardSubscribers.find(identifier);
	if(found != _keyboardSubscribers.end())
	{
	}*/
}


bool SchedulerExample::renderFrame(unsigned long elapsedMS)
{
	static float divideBy1000 = 0.001f;
	float elapsedS = static_cast<float>(elapsedMS) * divideBy1000;

	// modified from Ogre forums, post by nfz: http://www.ogre3d.org/phpBB2/viewtopic.php?t=6552&highlight=custom+render+loop
	FrameEvent frameEvent;
	frameEvent.timeSinceLastEvent = frameEvent.timeSinceLastFrame = elapsedS;

	// notify frame listeners that a frame is starting
	// break out of mainLoop if frame listener returns false
	_continueRendering = _root->_fireFrameStarted(frameEvent) && _continueRendering;

	_root->_updateAllRenderTargets();

	// notify frame listeners that the frame has ended
	// break out of mainLoop if frame listener returns false
	_continueRendering = _root->_fireFrameEnded(frameEvent) && _continueRendering;

	return true;
}

bool SchedulerExample::captureInput(unsigned long elapsedMS)
{
	_keyboard->capture();
	return true;
}

bool SchedulerExample::frameRate(unsigned long elapsedMS)
{
	float fps = 1000.0f / std::max(elapsedMS, 1UL);
	std::cout << fps << "\n";
	return true;
}

bool SchedulerExample::keyPressed(const OIS::KeyEvent& arg)
{
	if(arg.key == OIS::KC_ESCAPE)
	{
		_continueRendering = false;
	}

	for(KeyboardSubscriberMap::iterator i = _keyboardSubscribers.begin(), end = _keyboardSubscribers.end(); i != end; ++i)
	{
		i->second(arg);
	}
	return true;
}

bool SchedulerExample::keyReleased(const OIS::KeyEvent& arg)
{
	return true;
}

