

#include "SwirlingBlobbies.h"
#include "SchedulerExample.h"

#include "SimpleTime.h"


void delay(unsigned long ms)
{
	unsigned long timeout = Time::getMilliseconds() + ms;
	while(Time::getMilliseconds() < timeout);
}


ColourValue happyColour = ColourValue::Green;
ColourValue sadColour = ColourValue::Red;
ColourValue happinessDeltaColour = happyColour - sadColour;

const unsigned long blobUpdatePeriod = 20;


SwirlingBlob::SwirlingBlob(BillboardSet* bbset, float size, unsigned long variance, float radius, Radian initialAngle) :
	_bbset(bbset),
	_blob(_bbset->createBillboard(Vector3::ZERO)),
	_radius(radius),
	_angle(initialAngle)
{
	_blob->setDimensions(size, size);
	updatePosition(0);

	if(SwirlingBlobbies::getSingleton().getMovingAverageSmoothing())
	{
		_taskId = SchedulerExample::getSingleton().getMovingAverageTaskQueue().addTask(MakeDelegate(this, &SwirlingBlob::updatePosition), blobUpdatePeriod, variance);
	}
	else
	{
		_taskId = SchedulerExample::getSingleton().getFixedLimitTaskQueue().addTask(MakeDelegate(this, &SwirlingBlob::updatePosition), blobUpdatePeriod, variance);
	}
}

SwirlingBlob::~SwirlingBlob()
{
	_bbset->removeBillboard(_blob);

	if(SwirlingBlobbies::getSingleton().getMovingAverageSmoothing())
	{
		SchedulerExample::getSingleton().getMovingAverageTaskQueue().removeTask(_taskId);
	}
	else
	{
		SchedulerExample::getSingleton().getFixedLimitTaskQueue().removeTask(_taskId);
	}
}

bool SwirlingBlob::updatePosition(unsigned long elapsedMS)
{
	// all blobs rotate at a constant angular velocity
	static Radian rotationSpeed(0.4f);

	float elapsedS = 0.001f * blobUpdatePeriod;//elapsedMS;
	Radian rotationIncrement = rotationSpeed * elapsedS;
	_angle += rotationIncrement;

	Vector3 position = Quaternion(_angle, Vector3::UNIT_Z).xAxis() * _radius;
	_blob->setPosition(position);

	unsigned long extraDelay = std::max(blobUpdatePeriod, elapsedMS) - blobUpdatePeriod;
	float happiness = 1.0f - std::min(1.0f, (0.1f * extraDelay)); // 10ms is sad, 0ms is happy
	_blob->setColour(sadColour + (happinessDeltaColour * happiness));

	delay(2);

	return true;
}




template<> SwirlingBlobbies* Singleton<SwirlingBlobbies>::ms_Singleton = 0;


SwirlingBlobbies::SwirlingBlobbies(unsigned int initialBlobs) :
	_sceneNode(SchedulerExample::getSingleton().getSceneManager()->getRootSceneNode()->createChildSceneNode("Don't create multiple SwirlingBlobbies objects")),
	_blobSet(SchedulerExample::getSingleton().getSceneManager()->createBillboardSet("Don't create multiple SwirlingBlobbies objects", initialBlobs * 3)),
	_nextRadius(0.2f),
	_movingAverageSmoothing(false)
{
	_sceneNode->attachObject(_blobSet);

	_blobSet->setMaterialName("Blob");

	increaseBlobNumbers(initialBlobs);

	SchedulerExample::getSingleton().subscribeKeyboard("SwirlingBlobbies", MakeDelegate(this, SwirlingBlobbies::keyboardCallback));
}

SwirlingBlobbies::~SwirlingBlobbies()
{
	decreaseBlobNumbers(0);
}

void SwirlingBlobbies::keyboardCallback(const OIS::KeyEvent& arg)
{
	if(arg.key == OIS::KC_EQUALS)
	{
		increaseBlobNumbers(_largeBlobs.size() + 1);
	}
	else if(arg.key == OIS::KC_MINUS)
	{
		decreaseBlobNumbers(std::max<int>(0, _largeBlobs.size() - 1));
	}
	else if(arg.key == OIS::KC_SPACE)
	{
		toggleSmoothingMethod();
	}
}

void SwirlingBlobbies::increaseBlobNumbers(unsigned int number)
{
	// when blobs increase, restart all blobs again
	decreaseBlobNumbers(0);

	while(_largeBlobs.size() < number)
	{
		_largeBlobs.push_back(new SwirlingBlob(_blobSet, 0.2f, 0, _nextRadius, Radian(0.0f)));

		_mediumBlobs.push_back(new SwirlingBlob(_blobSet, 0.1f, 4, _nextRadius, Radian(0.0f)));

		_smallBlobs.push_back(new SwirlingBlob(_blobSet, 0.05f, 8, _nextRadius, Radian(0.0f)));

		_nextRadius += 0.2f;
	}
}

void SwirlingBlobbies::decreaseBlobNumbers(unsigned int number)
{
	while(_largeBlobs.size() > number)
	{
		delete _largeBlobs.back();
		_largeBlobs.pop_back();

		delete _mediumBlobs.back();
		_mediumBlobs.pop_back();

		delete _smallBlobs.back();
		_smallBlobs.pop_back();

		_nextRadius -= 0.2f;
	}
}

void SwirlingBlobbies::toggleSmoothingMethod()
{
	unsigned long blobs = _largeBlobs.size();
	decreaseBlobNumbers(0);

	_movingAverageSmoothing = !_movingAverageSmoothing;

	increaseBlobNumbers(blobs);
}

