// COPYRIGHT  2005, Paul Bridger. All rights reserved.
// Licensed under the new BSD license: see license.txt for details


#ifndef BOOST_SCHEDULER
#define BOOST_SCHEDULER


#include "TaskQueue.h"
#include "TaskOrderingPolicies.h"
#include "StepSmoothingPolicies.h"

#include <boost/function.hpp>


namespace Scheduler
{
	// CallbackType must return a bool indicating whether it should remain in the task queue,
	//  and should take an unsigned long -- the number of MS elapsed since the last call to that callback
	typedef boost::function<bool (unsigned long)> BoostCallback;


	/** Frame-locked task queue.

	Useful for tasks that must execute in sync with each other. 
	Eg. graphics rendering, input handling, possibly physics, etc.
	*/
	typedef TaskQueue<PriorityOrdering<BoostCallback>, NoSmoothing> FrameLockedQueue;

	/** Fixed-length frame queue, for periodic tasks.

	Useful for a set of tasks where each processing step should only take a given amount of time,
	and where low-priority tasks can tolerate being called at a much lower than expected frequency
	under periods of heavy load.
	(Can help create a fixed frame-rate).
	*/
	typedef TaskQueue<UrgencyOrdering<BoostCallback>, FixedLimitSmoothing> FixedLimitQueue;

	/** Adaptively adjusted frame-length queue, for periodic tasks.

	Useful for managing tasks under fluctuating load. As load increases, higher variance tasks are 
	used as a buffer to ensure low variance tasks are executed regularly. If load stays high, the
	over-capacity processing burden is spread amongst all tasks and frame rate drops.
	*/
	typedef TaskQueue<UrgencyOrdering<BoostCallback>, MovingAverageSmoothing> MovingAverageQueue;
}


#endif
