// COPYRIGHT  2005, Paul Bridger. All rights reserved.
// Licensed under the new BSD license: see license.txt for details


/** \mainpage
<h2>Basics</h2>
This library is built around the concept of a ::TaskQueue - a container to manage and call groups of tasks.

::TaskQueue's are customisable via policy classes, template parameters and other means, which give you a fair amount of 
flexibility in specifying how a given task queue will operate.

When creating a TaskQueue, you must supply template arguments to specify the frame-length smoothing policy 
(if any) and also the task ordering policy.

Policy-based design can be confusing at first, so a good place to start is with the predefined TaskQueue types:
- ::FrameLockedQueue
- ::FixedLimitQueue
- ::MovingAverageQueue

In my code, I find myself using a ::FrameLockedQueue for rendering and input handling tasks. I then throw all
my periodically-called tasks into a ::MovingAverageQueue. 

The ::FixedLimitQueue could be useful when you have a subset of tasks you want to allocate a resticted amount
of time to. (Or if you want your app to have a fairly strict maximum frame length).

<h2>CallbackType</h2>

New in this release is the ability to use a user defined CallbackType, rather than FastDelegate being hard-coded.
This is done by specifying the CallbackType template parameter on a TaskOrderingPolicy. 

The major upshot of this
is that you can now use (eg.) boost::function or your own callback type. The only (?) restriction on CallbackType
is that it is callable with an unsigned long and returns a boolean. 

For example, it may define: <code>bool operator()(unsigned long elapsedMs)</code>

The default CallbackType for this scheduler remains FastDelegate, but a separate include is provided so you can easily 
use boost::function.

Using the scheduler with boost::function opens the door for Python support, which is also included in this release.

<h2>Synopsis</h2>

<h3>C++</h3>
\code
Time::initialise();

// set up task queues<br>
FrameLockedQueue frameLockedQueue;

FixedLimitQueue fixedLimitQueue;<br>
fixedLimitQueue.setFrameLimit(20);

// add tasks
frameLockedQueue.addTask(&renderFrame, 1.0f);

fixedLimitQueue.addTask(&aiTask, 100, 20);
fixedLimitQueue.addTask(&pathfindingTask, 200, 20);

// if using FastDelegate
fixedLimitQueue.addTask(MakeDelegate(animationObject, &AnimationType::animationTaskMember), 20, 0);

// if using boost::function
fixedLimitQueue.addTask(boost::bind(&AnimationType::animationTaskMember, animationObject, _1), 20, 0);

// ...

// main loop
unsigned long lastMS = Time::getMilliseconds();
while(1)
{
	// update time, find elapsed time since last frame
	unsigned long nowMS = Time::getMilliseconds();
	unsigned long elapsedMS = nowMS - lastMS;
	lastMS = nowMS;

	// run frame-locked tasks
	frameLockedQueue.step(elapsedMS);

	// run variable tasks
	fixedLimitQueue.step(elapsedMS);
}
\endcode

<h3>Python</h3>
\verbatim
import Scheduler

print 'Module contents:'
print dir(Scheduler)

Scheduler.initialise()

frameQueue = Scheduler.createFrameLockedQueue()
limitQueue = Scheduler.createFixedLimitQueue()

def reportFrame(elapsedMs):
    print '=== Frame ==='
    return True

def reportLimit(elapsedMs):
    print 'hi'
    return True

frameQueue.addTask(reportFrame, 0.0)

limitQueue.addTask(reportLimit, 5, 0)
limitQueue.addTask(reportLimit, 10, 10)

for i in range(20):
    frameQueue.step(1)
    limitQueue.step(1)

\endverbatim
*/


#ifndef SCHEDULER
#define SCHEDULER


#include "TaskQueue.h"
#include "TaskOrderingPolicies.h"
#include "StepSmoothingPolicies.h"

#include "FastDelegate.h"


namespace Scheduler
{
	// CallbackType must return a bool indicating whether it should remain in the task queue,
	//  and should take an unsigned long -- the number of MS elapsed since the last call to that callback
	typedef fastdelegate::FastDelegate1<unsigned long, bool> FastDelegateCallback;


	/** Frame-locked task queue.

	Useful for tasks that must execute in sync with each other. 
	Eg. graphics rendering, input handling, possibly physics, etc.
	*/
	typedef TaskQueue<PriorityOrdering<FastDelegateCallback>, NoSmoothing> FrameLockedQueue;

	/** Fixed-length frame queue, for periodic tasks.

	Useful for a set of tasks where each processing step should only take a given amount of time,
	and where low-priority tasks can tolerate being called at a much lower than expected frequency
	under periods of heavy load.
	(Can help create a fixed frame-rate).
	*/
	typedef TaskQueue<UrgencyOrdering<FastDelegateCallback>, FixedLimitSmoothing> FixedLimitQueue;

	/** Adaptively adjusted frame-length queue, for periodic tasks.

	Useful for managing tasks under fluctuating load. As load increases, higher variance tasks are 
	used as a buffer to ensure low variance tasks are executed regularly. If load stays high, the
	over-capacity processing burden is spread amongst all tasks and frame rate drops.
	*/
	typedef TaskQueue<UrgencyOrdering<FastDelegateCallback>, MovingAverageSmoothing> MovingAverageQueue;
}


#endif
