// COPYRIGHT  2005, Paul Bridger. All rights reserved.
// Licensed under the new BSD license: see license.txt for details


#ifndef SCHEDULER_STEP_SMOOTHING_POLICIES
#define SCHEDULER_STEP_SMOOTHING_POLICIES


#include "MovingAverage.h"


namespace Scheduler
{
	/** Does no frame time smoothing.	*/
	class NoSmoothing
	{
		protected:
			/**
			We have a protected non-virtual destructor to prevent host class (::TaskQueue) users from
			converting a ::TaskQueue to this policy class and then deleting it
			(this would be undefined behaviour without a virtual destructor, 
			but we do not want a virtual destructor for performance reasons)
			*/
			~NoSmoothing() {}

			void frameStart();
			void frameEnd();
			bool continueFrame();
	};

	/** Ends frame processing at a fixed time.
	This smoothing method can be useful for ensuring a constant frame rate.
	*/
	class FixedLimitSmoothing
	{
		public:
			/// Set frame-length limit in milliseconds
			void setFrameLimit(unsigned long limitMS);

		protected:
			~FixedLimitSmoothing() {}

			void frameStart();
			void frameEnd();
			bool continueFrame();


			unsigned long _frameLimit;
			unsigned long _frameTimeout;
	};


	/** Uses a moving average frame-length to restrict frame processing.

	This smoothing method keeps a record of the length of previous frames, 
	and stops frame processing when the current frame exceeds this length
	by a configurable proportion.

	This smoothing method allows the system to maintain a steady frame rate if a short-term spike 
	in processing occurs, but will allow frame time to increase if processing load increases chronically (long-term).
	*/
	class MovingAverageSmoothing
	{
		public:
			/// Set the number of frames to average when calculating the frame-length target.
			void setMovingAverageFrames(size_t frames);

			/** Set the maximum proportion of the moving average that the subsequent frame can be.

			Eg. if the moving average frame length is 20ms, and the fluctuation proportion is 1.2,
			then the next frame will be restricted to 20 * 1.2 = 24ms in length
			*/
			void setFluctuationProportion(float fluctuation);

			/** Set the minimum considered frame-time.

			In a non-full task queue with this policy, there will be many processing frames 
			where nothing needs to be done and frame length is 0ms. If these are entered normally
			into the moving average, this provides a downward pressure on frame times which may be
			undesirable.

			This allows you to set a minimum target frame-length.

			Default is 10ms, for 100fps.
			*/
			void setMinFrameTime(unsigned long minFrameTimeMS);

	
		protected:
			MovingAverageSmoothing();
			~MovingAverageSmoothing() {}

			void frameStart();
			void frameEnd();
			bool continueFrame();


			Utility::MovingAverage<float> _averageFrameLength;
			unsigned long _frameStart;
			unsigned long _frameTimeout;
			float _maxFluctuationProportion;
			unsigned long _minFrameTimeMS;
	};
}


#endif