// COPYRIGHT  2005, Paul Bridger. All rights reserved.
// Licensed under the new BSD license: see license.txt for details

/*
	Changes by Sigfried McWild
	for some reason gcc doesn't recognize TaskContainer::iterator as a type automatically, so it needs typename (but only in the for loops for some reason)
	declared operator<() const for all the Task classes, it should have been like that already
*/

#ifndef SCHEDULER_TASK_ORDERING_POLICIES
#define SCHEDULER_TASK_ORDERING_POLICIES


#include <algorithm>
#include <queue>
#include <vector>
#include <cmath>


namespace Scheduler
{
	typedef unsigned int TaskId;

	/** Order tasks by a simple priority metric.

	With this task ordering policy, if a frame is cut short by the smoothing policy the uncalled tasks
	are *not* bumped to the front of the queue for the next processing frame. Every processing frame starts 
	with the highest priority tasks. So, don't use PriorityOrdering with anything other than NoSmoothing
	unless you have low-priority tasks that are okay to completely skip under heavy processing.

	Removing tasks from this type of queue is O(N) no matter what method you use:
	- the removeTask() method, or
	- the task returning false.
	*/
	template <typename CallbackType>
	class PriorityOrdering
	{
		public:
			/// Add a task to the queue.
			TaskId addTask(CallbackType callback, float priority)
			{
				_tasks.push_back(Task(_nextTaskId, priority, _timeMs, callback));
				std::sort(_tasks.begin(), _tasks.end());
				return _nextTaskId++;
			}

			/** Remove a task from the queue.
			
			This is O(N).
			*/
			void removeTask(TaskId id)
			{
				for(typename TaskContainer::iterator i = _tasks.begin(), end = _tasks.end(); i != end; ++i) //changed
				{
					if(i->id == id)
					{
						_tasks.erase(i);
						break;
					}
				}
			}


		protected:
			struct Task
			{
				Task(TaskId _id, float _priority, unsigned long _lastCallTime, CallbackType _callback) :
					id(_id), priority(_priority), lastCallTime(_lastCallTime), callback(_callback)
				{}

				// should rhs be before us?
				//  sort highest-priority first
				bool operator<(const Task& rhs) const  //changed
				{
					return (priority < rhs.priority);
				}

				TaskId id;
				float priority;
				unsigned long lastCallTime;
				CallbackType callback;
			};
			typedef std::vector<Task> TaskContainer;


		protected:
			PriorityOrdering() :
				_timeMs(0),
				_nextTask(0),
				_nextTaskId(0)
			{}

			/**
			We have a protected non-virtual destructor to prevent host class (::TaskQueue) users from
			converting a ::TaskQueue to this policy class and then deleting it
			(this would be undefined behaviour without a virtual destructor, 
			but we do not want a virtual destructor for performance reasons)
			*/
			~PriorityOrdering() {}

			// make next task the first task
			void reset()
			{
				_nextTask = 0;
			}

			// return whether any tasks remain to be processed
			bool runNextTask()
			{
				if(_tasks.empty() == false)
				{
					Task& t(_tasks[_nextTask]);

					// call task with elapsedMs
					if(t.callback(_timeMs - t.lastCallTime) == false)
					{
						removeTask(t.id);
					}
					else
					{
						t.lastCallTime = _timeMs;
					}

					if(++_nextTask == _tasks.size())
					{
						reset();
						return false;
					}
					else
					{
						return true;
					}
				}
				else
				{
					reset();
					return false;
				}
			}

			void taskOrderingStep(unsigned long sinceLastMs)
			{
				_timeMs += sinceLastMs;
			}


		protected:
			unsigned long _timeMs;

			TaskContainer _tasks;

			size_t _nextTask;

			TaskId _nextTaskId;
	};


	/** Order tasks based on a window of possible execution.

	Tasks are defined by an optimal calling period (time between calls, or 1/frequency), and by a variance value.
	These parameters define a window within which a task would like to be called. The next task to be called is
	always the one whose processing window ends first.

	As a result, under heavy processing load high variance tasks are deferred and will have an effective period
	somewhere between the optimal period and (optimal period + variance).

	Removing a task using the removeTask() method is O(N), but removing a task by having the task return false
	has zero cost.
	*/
	template <typename CallbackType>
	class UrgencyOrdering
	{
		public:
			/** Add a task to the queue.
			@param periodMs Optimal calling period in milliseconds.
			@param varianceMs Maximum calling variance in milliseconds.
			*/
			TaskId addTask(CallbackType callback, unsigned long periodMs, unsigned long varianceMs = 0)
			{
				_tasks.push_back(Task(_nextTaskId, _timeMs + periodMs, _timeMs, periodMs, varianceMs, callback));
				std::push_heap(_tasks.begin(), _tasks.end());
				return _nextTaskId++;
			}

			/** Remove a task.

			Removing a task using the removeTask() method is O(N), but removing a task by having the task return false
			has zero cost.*/
			void removeTask(TaskId id)
			{
				for(typename TaskContainer::iterator i = _tasks.begin(), end = _tasks.end(); i != end; ++i) //changed
				{
					if(i->id == id)
					{
						_tasks.erase(i);
						break;
					}
				}
				std::make_heap(_tasks.begin(), _tasks.end());

				for(typename ShortlistTaskContainer::iterator i = _taskShortlist.begin(), end = _taskShortlist.end(); i != end; ++i) //changed
				{
					if(i->id == id)
					{
						_taskShortlist.erase(i);
						break;
					}
				}
				std::make_heap(_taskShortlist.begin(), _taskShortlist.end());
			}


		protected:
			struct Task
			{
				Task(TaskId _id, unsigned long _earliestPossible, unsigned long _lastCallTime, unsigned long _period, unsigned long _variance, CallbackType _callback) :
					id(_id), earliestPossible(_earliestPossible), lastCallTime(_lastCallTime), period(_period), variance(_variance), callback(_callback)
				{}

				// should rhs be before us?
				//  for heap, sort earliest-possible first
				bool operator<(const Task& rhs) const //changed
				{
					return (earliestPossible > rhs.earliestPossible);
				}

				TaskId id;
				unsigned long earliestPossible, lastCallTime, period, variance;
				CallbackType callback;
			};

			struct ShortlistTask
			{
				ShortlistTask(TaskId _id, unsigned long _latestPossible, unsigned long _lastCallTime, unsigned long _period, unsigned long _variance, CallbackType _callback) :
					id(_id), latestPossible(_latestPossible), lastCallTime(_lastCallTime), period(_period), variance(_variance), callback(_callback)
				{}

				// should rhs be before us?
				//  for heap, sort latest-possible last
				bool operator<(const ShortlistTask& rhs) const  //changed
				{
					return (latestPossible > rhs.latestPossible);
				}

				TaskId id;
				unsigned long latestPossible, lastCallTime, period, variance;
				CallbackType callback;
			};

		protected:
			UrgencyOrdering() :
				_timeMs(0),
				_nextTaskId(0)
			{}

			~UrgencyOrdering() {}

			void addTask(const ShortlistTask& task)
			{
				// calculate earliest possible call time
				_tasks.push_back(Task(task.id, task.lastCallTime + task.period/*task.latestPossible*/, task.lastCallTime, task.period, task.variance, task.callback));
				std::push_heap(_tasks.begin(), _tasks.end());
			}

			void reset()
			{}

			bool runNextTask()
			{
				if(_taskShortlist.empty() == false)
				{
					std::pop_heap(_taskShortlist.begin(), _taskShortlist.end());
					ShortlistTask task = _taskShortlist.back();
					_taskShortlist.pop_back();

					// call task with elapsedMs,
					//  only add task back into normal task pool if task returns true
					if(task.callback(_timeMs - task.lastCallTime) == true)
					{
						task.lastCallTime = _timeMs;
						addTask(task);
					}

					return _taskShortlist.empty() == false;
				}
				else
				{
					return false;
				}
			}

			void taskOrderingStep(unsigned long sinceLastMs)
			{
				_timeMs += sinceLastMs;

				// update shortlist
				while(_tasks.empty() == false)
				{
					if(_tasks.front().earliestPossible > _timeMs)
					{
						break;
					}
					else
					{
						Task& task = _tasks.front();
						ShortlistTask shortlistTask(task.id, task.earliestPossible, task.lastCallTime, task.period, task.variance, task.callback);
						std::pop_heap(_tasks.begin(), _tasks.end());
						_tasks.pop_back();

						// we initialised with the task.earliestPossible, so now add variance to get latestPossible
						shortlistTask.latestPossible += shortlistTask.variance;
						_taskShortlist.push_back(shortlistTask);
						std::push_heap(_taskShortlist.begin(), _taskShortlist.end());
					}
				}
			}


		protected:
			unsigned long _timeMs;

			// contains all tasks, sorted earliest-possible first
			typedef std::vector<Task> TaskContainer;
			TaskContainer _tasks;

			// contains tasks that could be run anytime from now until their "latest possible" time
			//  sorted by increasing "lastest possible" (ie. urgency)
			typedef std::vector<ShortlistTask> ShortlistTaskContainer;
			ShortlistTaskContainer _taskShortlist;

			TaskId _nextTaskId;
	};
}

#endif