// COPYRIGHT  2005, Paul Bridger. All rights reserved.
// Licensed under the new BSD license: see license.txt for details



#include "StepSmoothingPolicies.h"

#include "MovingAverage.h"
#include "SimpleTime.h"

#include <cassert>
#include <cmath>


namespace Scheduler
{
	void NoSmoothing::frameStart()
	{}

	void NoSmoothing::frameEnd()
	{}

	bool NoSmoothing::continueFrame()
	{
		return true;
	}






	void FixedLimitSmoothing::setFrameLimit(unsigned long limitMS)
	{
		_frameLimit = limitMS;
	}

	void FixedLimitSmoothing::frameStart()
	{
		_frameTimeout = Time::getMilliseconds() + _frameLimit;
	}

	void FixedLimitSmoothing::frameEnd()
	{}

	bool FixedLimitSmoothing::continueFrame()
	{
		return (_frameTimeout > Time::getMilliseconds());
	}




	void MovingAverageSmoothing::setMovingAverageFrames(size_t frames)
	{
		_averageFrameLength.setMaxElements(frames);
	}

	void MovingAverageSmoothing::setMinFrameTime(unsigned long minFrameTimeMS)
	{
		_minFrameTimeMS = minFrameTimeMS;
	}

	void MovingAverageSmoothing::setFluctuationProportion(float fluctuation)
	{
		assert(fluctuation >= 1.0f);
		_maxFluctuationProportion = fluctuation;
	}

	MovingAverageSmoothing::MovingAverageSmoothing() :
		_averageFrameLength(10),
		_maxFluctuationProportion(1.2f),
		_minFrameTimeMS(10) // for 100 fps
	{
		// gradually increase restrictions
		_averageFrameLength.push(1000);
	}

	void MovingAverageSmoothing::frameStart()
	{
		_frameStart = Time::getMilliseconds();
		unsigned long stepMS = static_cast<unsigned long>(std::ceil(static_cast<float>(_averageFrameLength.average()) * _maxFluctuationProportion));
		_frameTimeout = _frameStart + stepMS;
	}

	void MovingAverageSmoothing::frameEnd()
	{
		//  records the length of empty frames as _minFrameTime,
		//  so we do not ever have a target frame length below _minFrameTime
		_averageFrameLength.push(static_cast<float>(std::max(_minFrameTimeMS, Time::getMilliseconds() - _frameStart)));
	}

	bool MovingAverageSmoothing::continueFrame()
	{

		return (_frameTimeout > Time::getMilliseconds());
	}

}

